from pathlib import Path
from typing import Dict, List, Optional, Tuple

from pandas import isna, DataFrame, notna

from eis1600.markdown.markdown_patterns import PARAGRAPH_UID_TAG_PATTERN, PUNCTUATION
from eis1600.models.BiosPunctuationModel import BiosPunctuationModel
from eis1600.models.EventsPunctuationModel import EventsPunctuationModel
from eis1600.models.PoetryDetectionModel import PoetryDetectionModel
from eis1600.processing.preprocessing import get_yml_and_miu_df, tokenize_miu_text
from eis1600.processing.postprocessing import merge_tagslists_without_duplicates, reconstruct_miu_text_with_tags
from eis1600.repositories.repo import NEW_PARAGRAPHS_REPO_ERROR_LOG


def test_for_poetry(tokens: List[str], debug: Optional[bool] = False):
    return PoetryDetectionModel().predict_is_poetry(tokens, debug)


def remove_punctuation(df: DataFrame) -> DataFrame:
    sections, tokens, tags = [], [], []
    for section, token, tag in df.itertuples(index=False):
        if token in PUNCTUATION:
            if not isna(section):
                raise ValueError(
                    'First token in this section is punctuation:\n'
                    f'{section}\n'
                    'This is not allowed during the re-splitting routine.\n'
                    )
            elif not isna(tag):
                # Extend TAGS_LIST of prev token with tags from the punctuation token
                if not isna(tags[-1]):
                    tags[-1].extend(tag)
                else:
                    tags[-1] = tag

        else:
            sections.append(section)
            tokens.append(token)
            tags.append(tag)

    return DataFrame(zip(sections, tokens, tags), columns=['SECTIONS', 'TOKENS', 'TAGS_LISTS'])


def get_old_paragraphs(df: DataFrame) -> List[Tuple[str, List[str]]]:
    paragraphs = []

    tokens = []
    curr_section_type = None
    for section, token, tags in df.itertuples(index=False):
        if isna(section):
            if not isna(token):
                if tags:
                    if 'NEWLINE' in tags:
                        tokens.append('\n')
                    if 'HEMISTICH' in tags:
                        tokens.append('%~%')
                tokens.append(token)
        elif PARAGRAPH_UID_TAG_PATTERN.match(section):
            paragraphs.append((curr_section_type, tokens))
            curr_section_type = PARAGRAPH_UID_TAG_PATTERN.match(section).group('cat')
            tokens = [token]

    paragraphs.append((curr_section_type, tokens))
    return paragraphs[1:]  # First element is (None, '') due to the MIU header


def remove_original_paragraphs(old_paragraphs: List[Tuple[str, List[str]]]) -> \
        Tuple[List[Tuple[str, List[str]]], List[Dict]]:
    mergeable_paragraphs = []
    unsplitted = []
    poetry_test_res = []
    lines = []
    for cat, tokens in old_paragraphs:
        if cat == 'POETRY':
            curr_line = []
            for token in tokens:
                if token == '\n':
                    lines.append(curr_line)
                    curr_line = []
                else:
                    curr_line.append(token)
            lines.append(curr_line)
        else:
            lines.append(tokens)

    is_poetry = False
    curr_meter = None
    for line in lines:
        if 20 >= len(line) >= 3:
            is_poetry, res = test_for_poetry(line)
            poetry_test_res.append(res)
            if is_poetry:
                if '%~%' not in line:
                    line = ['%~%'] + line

                if curr_meter:
                    if curr_meter == res['label']:
                        mergeable_paragraphs.extend(['\n'] + line)
                    else:
                        unsplitted.append(('POETRY', mergeable_paragraphs))
                        mergeable_paragraphs = line
                        curr_meter = res['label']
                else:
                    unsplitted.append(('UNDEFINED', mergeable_paragraphs))
                    mergeable_paragraphs = line
                    curr_meter = res['label']
            else:
                is_poetry = False
                if curr_meter:
                    unsplitted.append(('POETRY', mergeable_paragraphs))
                    mergeable_paragraphs = line
                    curr_meter = None
                else:
                    mergeable_paragraphs.extend(line)

        else:
            is_poetry = False
            if curr_meter:
                unsplitted.append(('POETRY', mergeable_paragraphs))
                mergeable_paragraphs = []
                curr_meter = None

            mergeable_paragraphs.extend(line)

    if mergeable_paragraphs:
        if is_poetry:
            unsplitted.append(('POETRY', mergeable_paragraphs))
        else:
            unsplitted.append(('UNDEFINED', mergeable_paragraphs))

    return unsplitted, poetry_test_res


def split_by_model(tokens: List[str], is_bio: bool) -> List[Tuple[str, str]]:
    if is_bio:
        # TODO insert Bio Punctuation model after training
        # punctuation_predictions = BiosPunctuationModel().predict_sentence(tokens)
        punctuation_predictions = [None] * len(tokens)
    else:
        punctuation_predictions = EventsPunctuationModel().predict_sentence(tokens)
    text_with_punctuation = ''
    for t, p in zip(tokens, punctuation_predictions):
        text_with_punctuation += t + ' '
        if p:
            text_with_punctuation += p + ' '
    paragraphs = text_with_punctuation.split('PERIOD ')

    # TODO its not _ its COMMA etc
    return [('UNDEFINED', paragraph + 'PERIOD') if paragraph[-1] != '_' else ('UNDEFINED', paragraph)
            for paragraph in paragraphs if paragraph]


def redefine_paragraphs(uid: str, miu_as_text: str) -> Tuple[List[Dict], str]:
    yml_handler, df_original = get_yml_and_miu_df(miu_as_text)
    miu_header = df_original['SECTIONS'].iat[0]

    # Remove original punctuation and paragraphs
    df_new = remove_punctuation(df_original)
    old_paragraphs = get_old_paragraphs(df_new)
    unsplitted_text, poetry_test_res = remove_original_paragraphs(old_paragraphs)

    # Get new punctuation and paragraphs
    new_paragraphs = []
    for cat, unsplitted in unsplitted_text:
        if cat == 'UNDEFINED':
            new_paragraphs.extend(split_by_model(unsplitted, yml_handler.is_bio()))
        else:
            new_paragraphs.append((cat, ' '.join(unsplitted)))

    # And now put everything back together
    text_with_new_paragraphs = miu_header + '\n\n'
    for cat, paragraph in new_paragraphs:
        text_with_new_paragraphs += f'::{cat}::\n{paragraph}\n\n'

    # Get DataFrame of new punctuation and paragraphs
    zipped = tokenize_miu_text(text_with_new_paragraphs, simple_mARkdown=True)
    df_punctuation = DataFrame(zipped, columns=['SECTIONS', 'TOKENS', 'PUNCTUATION'])

    df_original['TOKENS'].mask(isna(df_original['TOKENS']), other='', inplace=True)

    data = []
    tmp = []
    idx_original = 0
    idx_punctuation = 1
    while idx_original < len(df_original):
        section = df_original['SECTIONS'].iat[idx_original]
        token = df_original['TOKENS'].iat[idx_original]
        tags_list = df_original['TAGS_LISTS'].iat[idx_original]
        automated_punctuation = df_punctuation['PUNCTUATION'].iat[idx_punctuation]
        automated_section = df_punctuation['SECTIONS'].iat[idx_punctuation]

        tmp.append((token, token == df_punctuation['TOKENS'].iat[idx_punctuation], df_punctuation['TOKENS'].iat[idx_punctuation], idx_punctuation))
        if idx_original == 0:
            data.append((section, token, tags_list, None))
            idx_original += 1
        elif notna(token) and token == df_punctuation['TOKENS'].iat[idx_punctuation]:
            # Token in df_original and df_punctuation are the same
            data.append((automated_section, token, tags_list, automated_punctuation))
            idx_original += 1
            idx_punctuation += 1
        elif notna(token) and token in PUNCTUATION or token == '':
            # Token in df_original and df_punctuation are not the same because the original token was punctuation or
            # end of a paragraph which got removed in the new version
            data.append((section, token, tags_list, None))
            idx_original += 1
        elif notna(token) and df_punctuation['TOKENS'].iat[idx_punctuation] == '':
            # Token in df_original and df_punctuation are not the same because in the new version we arrived at the
            # end of a paragraph which was not there in the original version
            data.append((automated_section, '', None, automated_punctuation))
            idx_punctuation += 1
        else:
            print(token, df_punctuation['TOKENS'].iat[idx_punctuation])
            Path(NEW_PARAGRAPHS_REPO_ERROR_LOG).mkdir(exist_ok=True, parents=True)
            df_original.to_csv(NEW_PARAGRAPHS_REPO_ERROR_LOG + f'{uid}_original.csv')
            df_punctuation.to_csv(NEW_PARAGRAPHS_REPO_ERROR_LOG + f'{uid}_punc.csv')
            DataFrame(tmp, columns=['token', 'equals', 'punc', 'idx_punctuation']).to_csv(
                    NEW_PARAGRAPHS_REPO_ERROR_LOG + f'Footnotes_noise_example.{uid}_tmp.csv'
            )
            DataFrame(data, columns=['SECTIONS', 'TOKENS', 'TAGS_LISTS', 'PUNCTUATION']).to_csv(
                    NEW_PARAGRAPHS_REPO_ERROR_LOG + f'{uid}_new.csv'
            )
            error = f'Something in the alignment broke, check the DataFrames in {NEW_PARAGRAPHS_REPO_ERROR_LOG}'
            raise IndexError(error)

    df_new = DataFrame(data, columns=['SECTIONS', 'TOKENS', 'TAGS_LISTS', 'PUNCTUATION'])
    df_new['TAGS_LISTS'] = df_new[['TAGS_LISTS', 'PUNCTUATION']].apply(merge_tagslists_without_duplicates,
                                                                       key='PUNCTUATION', axis=1)

    updated_text = reconstruct_miu_text_with_tags(df_new[['SECTIONS', 'TOKENS', 'TAGS_LISTS']])

    return poetry_test_res, updated_text



from typing import List, Optional

from eis1600.models.Model import Model
from eis1600.helper.Singleton import singleton


@singleton
class BiosPunctuationModel(Model):

    def __init__(self) -> None:
        super().__init__('EIS1600_Pretrained_Models/camelbert-ca-bios-punctuation/')

    def predict_sentence(self, tokens: List[str], debug: Optional[bool] = False) -> List[str]:
        class2punct = {"B-COMMA": "،", "B-PERIOD": ".", "B-COLON": ":"}

        preds_class = super().predict_sentence(tokens, debug)
        preds_punct = [class2punct[p] if p in class2punct.keys() else None for p in preds_class]
        print(preds_punct)

        return preds_punct



from glob import glob
from typing import List, Union

from tqdm import tqdm

from eis1600.nlp.utils import insert_onom_tag, insert_onomastic_tags
from eis1600.processing.postprocessing import write_updated_miu_to_file
from eis1600.processing.preprocessing import get_yml_and_miu_df
from eis1600.training_data.online_editor_files import fix_formatting


def remove_nasab_tag(tags_list: Union[None, List[str]]):
    if tags_list and 'NASAB' in tags_list:
        tags_list.remove('NASAB')
    return tags_list


def annotation(path: str):
    with open(path, 'r', encoding='utf-8') as miu_file_object:
        # 1. open miu file and disassemble the file to its parts
        yml_handler, df = get_yml_and_miu_df(miu_file_object)

        df['ONOM_TAGS'] = insert_onom_tag(df)
        df['ONOMASTIC_TAGS'] = insert_onomastic_tags(df)
        df['TAGS_LISTS'] = df['TAGS_LISTS'].apply(remove_nasab_tag)

    with open(path.replace('12k/', '12k_copy/'), 'w', encoding='utf-8') as out_file_object:
        write_updated_miu_to_file(
                out_file_object, yml_handler, df[['SECTIONS', 'TOKENS', 'TAGS_LISTS', 'ONOM_TAGS', 'ONOMASTIC_TAGS']],
                forced_re_annotation=True
        )


def main():
    infiles = glob('12k/*.EIS1600')

    x = 869
    for idx, file in tqdm(list(enumerate(infiles[x:]))):
        try:
            fix_formatting(file)
        except ValueError:
            print(idx + x, file)
        except Exception as e:
            print(idx + x, file)
            print(e)

        annotation(file)



from logging import ERROR, Formatter
from sys import argv, exit
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from eis1600.helper.CheckFileEndingActions import CheckFileEndingEIS1600OrEIS1600TMPAction
from p_tqdm import p_uimap
from tqdm import tqdm

from eis1600.helper.logging import setup_logger
from eis1600.repositories.repo import TEXT_REPO, get_ready_and_double_checked_files
from eis1600.texts_to_mius.subid_methods import add_ids


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to insert UIDs in EIS1600TMP file(s) and thereby converting them to final EIS1600 
            file(s).
-----
Give a single EIS1600 or EIS1600TMP file as input.

Run without input arg to batch process all double-checked and ready files from the OpenITI_EIS1600_Texts directory.
'''
            )
    arg_parser.add_argument('-D', '--debug', action='store_true')
    arg_parser.add_argument(
            'input', type=str, nargs='?',
            help='EIS1600 or EIS1600TMP file to process',
            action=CheckFileEndingEIS1600OrEIS1600TMPAction
            )
    args = arg_parser.parse_args()

    infile = args.input
    debug = args.debug

    if infile:
        add_ids(infile)
    else:
        files_ready, files_double_checked = get_ready_and_double_checked_files()
        files = files_ready + files_double_checked

        if not files:
            print(
                    'There are no more EIS1600 files to process'
            )
            exit()

        print('\nAdd IDs')
        formatter = Formatter('%(message)s\n\n\n')
        logger = setup_logger('sub_ids', TEXT_REPO + 'sub_ids.log', ERROR, formatter)
        res = []
        count = 0
        if debug:
            x = 1
            for i, infile in tqdm(list(enumerate(files[x:]))):
                print(i + x, infile)
                try:
                    add_ids(infile)
                except ValueError as e:
                    logger.error(f'{infile}\n{e}\n\n')
                    count += 1

            print(f'{len(files)-count}/{len(files)} processed')
        else:
            res += p_uimap(add_ids, files)

    print('Done')






import re

from eis1600.markdown.markdown_patterns import AR_CHR, AR_STR

ABU = re.compile(r' [اأ]بو ')
ABI = re.compile(r' [اأ]ب[يى] ')
UMM = re.compile(r' [اأ]م ')
IBN_IBNA = re.compile(r' [اإ](بن[ةه]?)(?= )')
BN_BNT = re.compile(r' (بن[ت]?) ')
DIN_DAULA = re.compile(r' (الدين|الدول[ةه])')
BANU_BANI = re.compile(r'بن[ويى] ')
YURIFA_K_BI = re.compile(r'([تي]عرف) (ك' + AR_STR + r') (ب)')

CRF_PATTERN = re.compile(AR_CHR + r'*?عر' + AR_CHR + r'*?ف')
SHR_PATTERN = re.compile(AR_CHR + r'*?شت?ه' + AR_CHR + r'*?ر')
LQB_PATTERN = re.compile(AR_CHR + r'*?لق' + AR_CHR + r'*?ب')

LETTERS = r'(?:[أا]لف)|(?:باء)|(?:تاء)|(?:ثاء)|(?:جيم)|(?:حاء)|(?:خاء)|(?:داء)|(?:دال)|(?:ذاء)|(?:ذال)|(?:راء)|(' \
          r'?:زاء)|(?:زاي)|(?:زين)|(?:سين)|(?:شين)|(?:صاد)|(?:ضاد)|(?:طاء)|(?:ظاء)|(?:عين)|(?:غين)|(?:فاء)|(?:قاف)|(' \
          r'?:كاف)|(?:كاء)|(?:لام)|(?:ميم)|(?:نون)|(?:هاء)|(?:واو)|(?:ياء)|(?:همزة)|(?:تاء مربوطة)|(?:تاء ت[أا]نيث)|(' \
          r'?:[أا]لف مقصورة)|(?:تاء ثالثة)|(?:فتحة)|(?:فتح)|(?:مفتوحة)|(?:فتحتين)|(?:مفتوحتين)|(?:ضمة)|(?:ضم)|(' \
          r'?:مضمومة)|(?:ضمتين)|(?:مضمومتين)|(?:كسرة)|(?:كسر)|(?:مكسورة، كسرتين)|(?:مكسورتين)|(?:مدة)|(?:شدة)|(' \
          r'?:مشددة)|(?:تشديد)|(?:سكون)|(?:ساكنة)|(?:دون)|(?:ثم)|(?:بعدها)|(?:بينهما)|(?:أو)|(?:معجمة)|(?:حروف)|(' \
          r'?:آخر الحروف)|(?:مهملة)|(?:مهملتين)|(?:موحدة)|(?:موحدتين)|(?:مثناة)|(?:مثناتين)|(?:[أا]ولى)|(?:ثانية)|(' \
          r'?:ثالثة)|(?:في آخره)|(?:آخره)|(?:[أا]وله)|(?:تحتانية)|(?:تصريف)'

SPELLING = re.compile(r'(?<= )([بو]?(?:ال)?(?:' + LETTERS + r') ){2,}')



from glob import glob
from sys import argv
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from typing import List, Union
from re import compile

from eis1600.nlp.annotation_utils import insert_onom_tag, insert_onomastic_tags

from eis1600.markdown.EntityTags import EntityTags

from tqdm import tqdm

from eis1600.helper.CheckFileEndingActions import CheckIsDirAction
from eis1600.processing.postprocessing import write_updated_miu_to_file
from eis1600.processing.preprocessing import get_yml_and_miu_df
from eis1600.training_data.online_editor_files import fix_formatting

OLD_ONOMASTIC_TAGS_PATTERN = compile(r'Ü?(?:' + '|'.join(EntityTags().get_onom_tags()) + ')')
OLD_NER_TAGS_PATTERN = compile(r'Ü?(?:(?:P\d{1,2}[STFCOX]{0,2})|(?:T\d[BDKOPRX]))')


def remove_old_ner_tags(tags_list: Union[None, List[str]]) -> Union[None, List[str]]:
    if tags_list:
        kept_tags_str = OLD_NER_TAGS_PATTERN.sub(','.join(tags_list), '')
        return [t for t in kept_tags_str.split(',') if t]

    return tags_list


def remove_old_tags(tags_list: Union[None, List[str]], which_annotations: List[str]) \
        -> Union[None, List[str]]:
    if tags_list:
        kept_tags_str = ','.join(tags_list)
        if 'NER' in which_annotations:
            kept_tags_str = OLD_NER_TAGS_PATTERN.sub(kept_tags_str, '')
        if 'O' in which_annotations:
            kept_tags_str = OLD_ONOMASTIC_TAGS_PATTERN.sub(kept_tags_str, '')
        if 'P' in which_annotations:
            pass
        if 'T' in which_annotations:
            pass

        return [t for t in kept_tags_str.split(',') if t]

    return tags_list


def reannotation(path: str, which_annotations: List[str]):
    with open(path, 'r+', encoding='utf-8') as miu_file_object:
        # 1. open miu file and disassemble the file to its parts
        yml_handler, df = get_yml_and_miu_df(miu_file_object)
        df['TAGS_LISTS'] = df['TAGS_LISTS'].apply(remove_old_tags, which_annotations=which_annotations)

        if 'NER' in which_annotations:
            df = df
        if 'O' in which_annotations:
            df['ONOM_TAGS'] = insert_onom_tag(df)
            df['ONOMASTIC_TAGS'] = insert_onomastic_tags(df)
        if 'P' in which_annotations:
            df = df
        if 'T' in which_annotations:
            df = df

        write_updated_miu_to_file(
            miu_file_object, yml_handler, df[['SECTIONS', 'TOKENS', 'TAGS_LISTS', 'ONOM_TAGS', 'ONOMASTIC_TAGS']],
            forced_re_annotation=True
        )


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to re-annotated files from the online-editor.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')
    arg_parser.add_argument('-ner', '--NER', action='store_true')
    arg_parser.add_argument('-o', '--onomastics', action='store_true')
    arg_parser.add_argument('-p', '--persons', action='store_true')
    arg_parser.add_argument('-t', '--toponyms', action='store_true')
    arg_parser.add_argument(
            'input', type=str, nargs='1',
            help='Directory which holds the files to process or individual file to annotate',
            action=CheckIsDirAction
    )

    which_annotations = []

    args = arg_parser.parse_args()
    debug = args.debug
    if args.NER:
        which_annotations.append('NER')
    if args.onomastics:
        which_annotations.append('O')
    if args.persons:
        which_annotations.append('P')
    if args.toponyms:
        which_annotations.append('T')

    in_dir = args.input
    if in_dir[-1] != '/':
        in_dir += '/'
    infiles = glob(in_dir + '*.EIS1600')

    x = 0
    for idx, file in tqdm(list(enumerate(infiles))):
        try:
            # Needed for online editor files
            # TODO better this is run while files are exported from DB
            fix_formatting(file)
        except ValueError:
            print(idx + x, file)
        except Exception as e:
            print(idx + x, file)
            print(e)

        reannotation(file, which_annotations)



from re import compile

from eis1600.markdown.EntityTags import EntityTags

PUNCTUATION_DICT = {'.': 'PERIOD', ',': 'LATINCOMMA', '،': 'COMMA', ':': 'COLON'}
PUNCTUATION = PUNCTUATION_DICT.keys()

AR_LETTERS_CHARSET = frozenset(
        u'\u0621\u0622\u0623\u0624\u0625\u0626\u0627'
        u'\u0628\u0629\u062a\u062b\u062c\u062d\u062e'
        u'\u062f\u0630\u0631\u0632\u0633\u0634\u0635'
        u'\u0636\u0637\u0638\u0639\u063a\u0640\u0641'
        u'\u0642\u0643\u0644\u0645\u0646\u0647\u0648'
        u'\u0649\u064a\u0671\u067e\u0686\u06a4\u06af'
)
AR_CHR = r'[' + u''.join(AR_LETTERS_CHARSET) + ']'
AR_STR = AR_CHR + '+'
AR_STR_AND_TAGS = r'[' + u''.join(AR_LETTERS_CHARSET) + 'a-zA-ZÜ0-9]+'
WORD = r'(?:(^|\s)' + AR_STR + ')'
NOISE_ELEMENTS = r'(?: [\[\]0-9،.():~|-])*'
AR_CHR_AND_NOISE = r'[' + u''.join(AR_LETTERS_CHARSET) + r'0-9"({\[«-]'

# EIS1600 mARkdown
UID = r'(?P<UID>\d{12}(?:-\d{8})?)'
UID_WITHOUT_CAPTURING_GROUP = r'\d{12}(?:-\d{8})?'
UID_TAG = r'_ء_(#)?=' + UID + '= '
UID_TAG_PATTERN = compile(UID_TAG)
MIU_UID_TAG = r'_ء_#=(?P<UID>\d{12})= '
MIU_UID_TAG_WITHOUT_CAPTURING_GROUP = r'_ء_#=\d{12}= '
MIU_UID_TAG_PATTERN = compile(MIU_UID_TAG)
MIU_SPLITTER_PATTERN = compile(r'(?:^|\n\n)(?=' + MIU_UID_TAG + ')')
PARAGRAPH_CAT = '::(?P<cat>[A-Z_]+)::'
PARAGRAPH_CAT_PATTERN = compile(PARAGRAPH_CAT)
PARAGRAPH_CAT_WITHOUT_CAPTURING_GROUPS = '::[A-Z_]+::'
PARAGRAPH_SIMPLE_SPLITTER_PATTERN = compile('\n\n(' + PARAGRAPH_CAT_WITHOUT_CAPTURING_GROUPS + ')\n')
PARAGRAPH_UID_TAG = r'_ء_=' + UID_WITHOUT_CAPTURING_GROUP + '= ' + PARAGRAPH_CAT + ' ~'
PARAGRAPH_UID_TAG_WITHOUT_CAPTURING_GROUPS = r'_ء_=' + UID_WITHOUT_CAPTURING_GROUP + '= ' + PARAGRAPH_CAT_WITHOUT_CAPTURING_GROUPS + ' ~'
PARAGRAPH_UID_TAG_PATTERN = compile(PARAGRAPH_UID_TAG)
HEADER_END_PATTERN = compile(r'(#META#Header#End#)\n')
MIU_HEADER = r'#MIU#Header#'
MIU_HEADER_PATTERN = compile(MIU_HEADER)
HEADING_PATTERN = compile(UID_TAG + r'(?P<level>[|]+) (?P<heading>.*)\n')
PAGE_TAG = r' ?(?P<page_tag>PageV\d{2}P\d{3,})'
PAGE_TAG_PATTERN = compile(PAGE_TAG)
ONLY_PAGE_TAG = PARAGRAPH_UID_TAG + r'\n' + PAGE_TAG
ONLY_PAGE_TAG_PATTERN = compile(ONLY_PAGE_TAG)
PAGE_TAG_IN_BETWEEN_PATTERN = compile(
        AR_STR + r' ?' + r'\n\n' + ONLY_PAGE_TAG + r'\n\n' + PARAGRAPH_UID_TAG_WITHOUT_CAPTURING_GROUPS + ' ~\n' + AR_STR
)
TEXT_START_PATTERN = compile(MIU_UID_TAG + r'[|]')
SIMPLE_MARKDOWN_TEXT_START_PATTERN = compile(r'# [|]')
PARAGRAPH_TAG_MISSING = compile(r'(\n\n[^_])|(\n\n' + MIU_UID_TAG + r'[^\n]+\n(?:_ء_ )?)' + AR_CHR)
SIMPLE_MARKDOWN = compile(r'\n#')
SPAN_ELEMENTS = compile(r'</?span/?>')
POETRY_ATTACHED_AFTER_PAGE_TAG = compile('Page[VP0-9]+[^\n%]+%')

# MIU_TAG_PATTERN is used to split text - indices depend on the number of capturing groups so be careful when
# changing them
MIU_TAG_PATTERN = compile(r'(' + MIU_UID_TAG + r'(?P<category>[^\n]+))')
CATEGORY_PATTERN = compile(r'[$|@]+(?:[A-Z_]+[|$])?')
PARAGRAPH_SPLITTER_PATTERN = compile(r'\n\n(' + PARAGRAPH_UID_TAG_WITHOUT_CAPTURING_GROUPS + ')\n(?:_ء_)?')
TAG_PATTERN = compile(r'Ü?(?:[a-zA-Z_%~]+(?:\.[a-zA-Z0-9_%~]+)?)|' + PAGE_TAG + '|::|' +
                      '|'.join(PUNCTUATION_DICT.values()))
NOR_DIGIT_NOR_AR_STR = r'[^\d\n' + u''.join(AR_LETTERS_CHARSET) + ']*?'
TAG_AND_TEXT_SAME_LINE = r'([$@]+' + NOR_DIGIT_NOR_AR_STR + r'\d*' + NOR_DIGIT_NOR_AR_STR + r') ?((?:[(\[] ?)?' + AR_STR + r')'
MIU_UID_TAG_AND_TEXT_SAME_LINE_PATTERN = compile(r'(' + MIU_UID_TAG_WITHOUT_CAPTURING_GROUP + ')' + TAG_AND_TEXT_SAME_LINE)

# Catches MIU tags for BIO, CHR and PARATEXT, EDITOR, etc. (everything in between pipes).
# Does not catch HEADERS!
MIU_TAG_AND_TEXT_PATTERN = compile(r'(' + MIU_UID_TAG + r'(?:[$@]+?|\|[A-Z]+\|)(?: \d+)?)\n((?:\( ?)?' + AR_STR + r')')

# MIU entity tags
entity_tags = '|'.join(EntityTags().get_entity_tags())
ENTITY_TAGS_PATTERN = compile(r'\bÜ?(?P<full_tag>'
                              r'(?P<entity>' + entity_tags + r')(?P<length>\d{1,2})'
                                                             r'(?:(?P<sub_cat>[A-Z]+)|['r'A-Z0-9]+)?)\b')
YEAR_PATTERN = compile(r'Ü?Y(?P<num_tokens>\d{1,2})(?P<cat>[A-Z])(?P<written>\d{4}|None)(?P<i>I)?Y(?P<real>\d{4})?')
AGE_PATTERN = compile(r'Ü?A\d(?P<cat>[A-Z])(?P<written>\d{2,4})(?P<i>I)?A(?P<real>\d{2,4})?')
LUNAR_PATTERN = compile(r'Ü?L\d(?P<cat>X)(?P<written>\d{2})L')
TOPONYM_PATTERN = compile(r'Ü?T(?P<num_tokens>\d{1,2})(?P<cat>[A-Z])')
onom_tags = '|'.join(EntityTags().get_onom_tags())
ONOM_TAGS_PATTERN = compile(r'Ü?(?P<entity>' + onom_tags + r')(?P<length>\d{1,2})')

# EIS1600 light mARkdown
SIMPLE_HEADING_OR_BIO_PATTERN = compile(r'# [|$]+')
MIU_LIGHT_OR_EIS1600_PATTERN = compile(r'#|_ء_#')
PAGE_TAG_ON_NEWLINE_TMP_PATTERN = compile(r'(?<!\n)\n' + PAGE_TAG + r'(?=\n)')
SIMPLE_PARAGRAPH_PATTERN = compile(r'\n' + PARAGRAPH_CAT)

# Fix mARkdown files
SPACES_CROWD_PATTERN = compile(r'  +')
NEWLINES_CROWD_PATTERN = compile(r'\n{3,}')
NEW_LINE_BUT_NO_EMPTY_LINE_PATTERN = compile(r'[^\n]\n(?:(?:# [|$])|(?:' + UID_TAG + '))')
NEW_LINE_INSIDE_PARAGRAPH_NOT_POETRY_PATTERN = compile(r'(?<=\n)[^\n%~]+\n[^\n%]+\n')
TILDA_HICKUPS_PATTERN = compile(r'~\n~')
MISSING_DIRECTIONALITY_TAG_PATTERN = compile(r'(\n+)(' + AR_CHR_AND_NOISE + '|%~%|Page|ms)')
EMPTY_PARAGRAPH = r'::( ~)?\n(?!' + AR_CHR_AND_NOISE + '|%~%|Page|ms|_ء_)'
EMPTY_PARAGRAPH_CHECK_PATTERN = compile(EMPTY_PARAGRAPH)
SPACES_AFTER_NEWLINES_PATTERN = compile(r'\n +')
POETRY_PATTERN = compile(
        r'# (' + AR_STR_AND_TAGS + '(?: ' + AR_STR_AND_TAGS + ')* %~% ' + AR_STR_AND_TAGS + '(?: ' +
        AR_STR_AND_TAGS +
        r')*) ?'
)
BELONGS_TO_PREV_PARAGRAPH_PATTERN = compile(r'\n(.{1,10})\n')
PAGE_TAG_ON_NEWLINE_MARKDOWN_PATTERN = compile(r'\n' + PAGE_TAG)
PAGE_TAG_SPLITTING_PARAGRAPH_PATTERN = compile(
        '(' + AR_STR + ' ?)' + r'\n\n' + PAGE_TAG + r'\n\n' + '(' + AR_STR +
        ')'
)
NORMALIZE_BIO_CHR_MD_PATTERN = compile('# ([$@]((BIO|CHR)_[A-Z]+[$@])| RAW)')
BIO_CHR_TO_NEWLINE_PATTERN = compile(TAG_AND_TEXT_SAME_LINE)

# Fixed poetry old file path pattern
FIXED_POETRY_OLD_PATH_PATTERN = compile(r'/Users/romanov/_OpenITI/_main_corpus/\w+/data/')



from typing import List, Optional
from threading import Lock

from camel_tools.ner import NERecognizer


class Model:
    def __init__(self, model_path: str) -> None:
        self.model_path = model_path
        self.model = NERecognizer(model_path)
        self.lock = Lock()

    def predict_sentence(self, tokens: List[str], debug: Optional[bool] = False) -> List[str]:
        with self.lock:
            if debug:
                print(self.model_path)
            return self.model.predict_sentence(tokens)

    def predict_sentence_with_windowing(self, tokens: List[str]) -> List[str]:
        windows = []
        batch_length = 450
        window_size = 50

        i = 0
        # TODO give more overlap to last window because left-over string could be only 3 tokens long -> cannot be
        #  analysed
        last_window_overlap = 0
        while len(tokens) > i * batch_length:
            start = i * batch_length
            if i > 0:
                start -= window_size
            end = (i + 1) * batch_length
            if end >= len(tokens):
                end = len(tokens) - 1
            windows.append(tokens[start: end])
            i += 1

        print(tokens)
        for w in windows:
            print(w)

        prediction_lists = []
        with self.lock:
            for window in windows:
                prediction_lists.append(self.model.predict_sentence(window))

        # Align windows
        predictions = []
        for idx, pred_list in enumerate(prediction_lists):
            preds = pred_list[window_size:]
            if idx == 0:
                preds = pred_list
            predictions.extend(preds)

        print(f'Tokens: {len(tokens)}, Predictions: {len(predictions)}')
        return prediction_lists



from eis1600.markdown.markdown_patterns import MISSING_DIRECTIONALITY_TAG_PATTERN
from eis1600.texts_to_mius.subid_methods import pre_clean_text, update_ids
from eis1600.yml.yml_handling import extract_yml_header_and_text
from eis1600.yml.YAMLHandler import YAMLHandler


def fix_formatting(file: str):
    with open(file, 'r+', encoding='utf-8') as fh:
        yml_str, text = extract_yml_header_and_text(fh, False)
        yml_handler = YAMLHandler().from_yml_str(yml_str)

        # Files from the online editor are stript of directionality tags and paragraph tags for ease of annotation
        # 1. Reconstruct MIU tag
        updated_text = text.replace('#', '_ء_#')
        # 2. Some more pre_cleaning
        updated_text = pre_clean_text(updated_text)
        # 3. Add directionality tags at the beginning of a text line
        updated_text = MISSING_DIRECTIONALITY_TAG_PATTERN.sub('\g<1>_ء_ \g<2>', updated_text)
        # 4. Add paragraph tags
        updated_text = update_ids(updated_text)

        fh.seek(0)
        fh.write(str(yml_handler) + updated_text)
        fh.truncate()



from sys import argv
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from eis1600.corpus_analysis.analyse_all_on_cluster import routine_per_text
from eis1600.helper.CheckFileEndingActions import CheckFileEndingEIS1600TextAction


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to re-annotated files from the online-editor.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')
    arg_parser.add_argument('-P', '--parallel', action='store_true')
    arg_parser.add_argument(
            'input', type=str, nargs='?',
            help='EIS1600 text file to annotate.',
            action=CheckFileEndingEIS1600TextAction
    )
    args = arg_parser.parse_args()
    debug = args.debug
    parallel = args.parallel
    infile = args.input

    routine_per_text(infile, parallel, force=True, debug=debug)



from pathlib import Path
from sys import argv
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from typing import Dict, Tuple

from p_tqdm import p_uimap

from json import dump
from datetime import datetime
from numpy import nan
from pandas import notna

from eis1600.bio.md_to_bio import md_to_bio
from eis1600.markdown.markdown_patterns import TOPONYM_PATTERN
from eis1600.model_evaluations.EvalResultsEncoder import EvalResultsEncoder
from eis1600.model_evaluations.eval_date_model import eval_dates_entity_recognition_and_classification
from eis1600.depricated.disassemble_reassemble_methods import get_yml_and_miu_df
from eis1600.repositories.repo import TRAINING_DATA_REPO, TRAINING_RESULTS_REPO
from eis1600.toponyms.methods import toponym_category_annotate_miu
from eis1600.toponyms.toponym_categories import TOPONYM_CATEGORIES_REPLACEMENTS, TOPONYM_LABEL_DICT


def combine_tokens_and_tags(row) -> str:
    return row['TOPONYMS_TRUE'][:-1] + ' ' + row['TOKENS']


def remove_category(val: str) -> str:
    if notna(val):
        return val[:-1]
    else:
        return val


def reconstruct_automated_tag(row) -> str:
    return 'ÜT' + row['num_tokens'] + row['cat']


def get_dates_true_and_pred(file: str) -> Tuple[Dict, Dict]:
    """Get ground-truth and prediction on labels and numerical values for MIU.

    :param str file: file path for MIU file.
    :param Dict bio_dict: BIO labels dictionary.
    :return Tuple[DataFrame, Dict, Dict]: DataFrame year contains two columns ('true' and 'pred'), the other two are
    dictionaries, one with the BIO labels derived from the ground-truth and the other with the BIO labels based on
    the predictions.
    """
    with open(file, 'r', encoding='utf-8') as miu_file_object:
        yml_handler, df = get_yml_and_miu_df(miu_file_object)

    # Extract the ground-truth annotated with EIS1600 tags
    s_notna = df['TAGS_LISTS'].loc[df['TAGS_LISTS'].notna()].apply(lambda tag_list: ','.join(tag_list))
    df_true = s_notna.str.extract(TOPONYM_PATTERN).dropna(how='all')
    toponyms = df_true.apply(reconstruct_automated_tag, axis=1)
    toponyms.name = 'TOPONYMS_TRUE'

    if not toponyms.empty:
        df = df.join(toponyms)
    else:
        df['TOPONYMS_TRUE'] = nan

    # Predict dates based on the text as EIS1600 tags
    df['TOPONYMS_PRED'] = toponym_category_annotate_miu(df['TOKENS'], df['TOPONYMS_TRUE'].apply(remove_category))

    # Parse EIS1600 tags to BIO tags for predictions and ground-truth
    bio_true = md_to_bio(
            df[['TOKENS', 'TOPONYMS_TRUE']],
            'TOPONYMS_TRUE',
            TOPONYM_PATTERN,
            'TO',
            TOPONYM_LABEL_DICT,
            True,
            TOPONYM_CATEGORIES_REPLACEMENTS
    )
    bio_pred = md_to_bio(
            df[['TOKENS', 'TOPONYMS_PRED']],
            'TOPONYMS_PRED',
            TOPONYM_PATTERN,
            'TO',
            TOPONYM_LABEL_DICT,
            True,
            TOPONYM_CATEGORIES_REPLACEMENTS
    )

    return bio_true, bio_pred


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to annotate onomastic information in gold-standard MIUs.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')

    args = arg_parser.parse_args()
    debug = args.debug

    with open(TRAINING_DATA_REPO + '5k_gold_standard.txt', 'r', encoding='utf-8') as fh:
        files_txt = fh.read().splitlines()

    infiles = [TRAINING_DATA_REPO + '5k_gold_standard/' + file for file in files_txt if Path(
            TRAINING_DATA_REPO + '5k_gold_standard/' + file
    ).exists()]

    res = []
    if debug:
        for i, file in enumerate(infiles):
            print(i, file)
            res.append(get_dates_true_and_pred(file))
    else:
        res += p_uimap(get_dates_true_and_pred, infiles)

        truth, predictions = zip(*res)

        with open(TRAINING_DATA_REPO + 'toponyms_truth.json', 'w', encoding='utf-8') as fh:
            dump(truth, fh, indent=4, ensure_ascii=False)
        with open(TRAINING_DATA_REPO + 'toponyms_predictions.json', 'w', encoding='utf-8') as fh:
            dump(predictions, fh, indent=4, ensure_ascii=False)

    truth = []
    predictions = []

    for bio_true, bio_pred in res:
        truth.append(bio_true['ner_classes'])
        predictions.append(bio_pred['ner_classes'])

    all_metrics = {
            'model': 'toponyms-rule-based'
    }
    all_metrics.update(eval_dates_entity_recognition_and_classification(truth, predictions))

    with open(TRAINING_RESULTS_REPO + 'toponyms/result-' + datetime.now().strftime('%Y%m%d-%H%M%S') + '.json', 'w',
              encoding='utf-8') as fh:
        dump(all_metrics, fh, cls=EvalResultsEncoder)

    print('Done')



from typing import Dict, List, Optional, Set, Tuple, Union, Iterator

from itertools import combinations
from operator import itemgetter
from os import makedirs
from os.path import dirname, split, splitext

from camel_tools.utils.charsets import UNICODE_PUNCT_CHARSET
from pandas import DataFrame, notna

from eis1600.gazetteers.Toponyms import Toponyms
from eis1600.helper.logging import setup_logger
from eis1600.markdown.EntityTags import EntityTags
from eis1600.markdown.markdown_methods import get_yrs_tag_value
from eis1600.markdown.markdown_patterns import ENTITY_TAGS_PATTERN, MIU_HEADER_PATTERN, NEWLINES_CROWD_PATTERN
from eis1600.miu.HeadingTracker import HeadingTracker
from eis1600.repositories.repo import GAZETTEERS_REPO
from eis1600.yml.YAMLHandler import YAMLHandler

__log_filename_nasab = GAZETTEERS_REPO + 'logs/nasab_known.log'
makedirs(dirname(__log_filename_nasab), exist_ok=True)
LOGGER_NASAB_KNOWN = setup_logger('nasab_known', __log_filename_nasab)
LOGGER_TOPONYMS_UNKNOWN = setup_logger('toponyms_unknown', GAZETTEERS_REPO + 'logs/toponyms_unknown.log')


def create_yml_header(category: str, headings: Optional[HeadingTracker] = None) -> str:
    """Creates a YAML header for the current MIU file and returns it as yamlfied string.

    :param str category: Category of the entry.
    :param Type[HeadingsTracker] headings: HeadingTracker with the super elements of the current MIU, optional.
    :return str: YAML header for the current MIU.
    """
    yml_header = YAMLHandler()
    yml_header.set_category(category)
    if headings:
        yml_header.set_headings(headings)

    return yml_header.get_yamlfied()


def extract_yml_header_and_text(miu_text_line_iter: Iterator, is_header: Optional[bool] = False) -> (str, str):
    """ Returns the YAML header and the text as a tuple from MIU file object.

    Splits the MIU file into a tuple of YAML header and text.
    :param Iterator miu_text_line_iter: Iterator of lines over the MIU file from which to extract YAML header and text.
    :param bool is_header: Indicates if the current MIU is the YAML header of the whole work and if so skips
    removing blank lines, defaults to False.
    :return (str, str): Tuple of the extracted YAML header and text.
    """
    text = ''
    miu_yml_header = ''
    for line in miu_text_line_iter:
        if MIU_HEADER_PATTERN.match(line):
            # Omit the #MIU#Header# line as it is only needed inside the MIU.EIS1600 file, but not in YMLDATA.yml
            next(miu_text_line_iter)
            line = next(miu_text_line_iter)
            miu_yml_header = ''
            while not MIU_HEADER_PATTERN.match(line):
                miu_yml_header += line
                line = next(miu_text_line_iter)
            # Omit the empty line between the header content and the #MIU#Header# line
            miu_yml_header = miu_yml_header[:-2]
            # Skip empty line after #MIU#Header#
            next(miu_text_line_iter)
        else:
            text += line
        # Replace new lines which separate YAML header from text
        if not is_header:
            text = NEWLINES_CROWD_PATTERN.sub('\n\n', text)

    return miu_yml_header, text


def add_to_entities_dict(
        entities_dict: Dict,
        cat: str,
        entity: Union[str, Tuple[str, Union[int, str]], List[Tuple[str, str]], List[str], Tuple[int, str], Dict],
        tag: Optional[str] = None
) -> None:
    """Add a tagged entity to the respective list in the entities_dict.

    :param Dict entities_dict: Dict containing previous tagged entities.
    :param str cat: Category of the entity.
    :param Union[str|int] entity: Entity - might be int if entity is a date, otherwise str.
    :param str tag: Optional, onomastic classification used to differentiate between onomastic elements, defaults to
    None.
    """
    cat = cat.lower() + 's'
    if tag:
        tag = tag.lower()
    if cat in entities_dict.keys():
        if cat == 'onomastics' and tag:
            if tag in entities_dict[cat].keys():
                entities_dict[cat][tag].add(entity)
            else:
                entities_dict[cat][tag] = {entity}
        else:
            entities_dict[cat].append(entity)
    else:
        if cat == 'onomastics' and tag:
            entities_dict[cat] = {}
            entities_dict[cat][tag] = {entity}
        elif isinstance(entity, list):
            entities_dict[cat] = entity
        else:
            entities_dict[cat] = [entity]


def add_annotated_entities_to_yml(
        df: DataFrame,
        yml_handler: YAMLHandler,
        file_path: str,
        reconstructed_miu_text_with_tags: str
) -> None:
    """Populates YAMLHeader with annotated entities.

    Extract annotated entities as metadata. While doing so, identify toponyms and calculate active period for the
    biographee.
    :param DataFrame df: DataFrame with two columns 'TAGS_LISTS' and 'TOKENS'.
    :param YAMLHandler yml_handler: YAMLHandler of the MIU.
    :param str file_path: Filename of the current MIU (used in error msg).
    """
    # We do not need to differentiate between automated and manual tags
    entity_tags_df = EntityTags().get_entity_tags_df()
    entities_dict = {}
    nas_dict = {}
    nas_counter = 0
    settlements_set: Set[str] = set()
    provinces_set: Set[str] = set()
    ambiguous_toponyms = False

    # Get all tags with their information extracted into individual elements
    s_notna = df['TAGS_LISTS'].loc[df['TAGS_LISTS'].notna()].apply(lambda tag_list: ','.join(tag_list))
    df_matches = s_notna.str.extractall(ENTITY_TAGS_PATTERN).dropna(how='all')

    # Interpret extracted information for each tag
    for index_tuple, row in df_matches.iterrows():
        # Since there can be multiple tags for the same tokens df_matches has a multi-index
        index = index_tuple[0]
        tag = row['entity']
        length = int(row['length'])
        sub_cat = row['sub_cat']
        try:
            entity = ' '.join(df['TOKENS'].iloc[index:index + length].to_list())
        except TypeError:
            print(f'Something is at odd here: {row["full_tag"]}\nCheck: {file_path}')
            yml_handler.set_error_while_collecting_annotated_entities(row["full_tag"])
            return
        cat = entity_tags_df.loc[entity_tags_df['TAG'].str.fullmatch(tag), 'CATEGORY'].iloc[0]

        if cat == 'DATE' or cat == 'LUNAR_MONTH' or cat == 'AGE':
            # TODO what about the month? ATM, they are not collected into the yml. If you want to change that,
            #  add 'month' (or maybe lunar_month) to YAMLHandler line 27!
            try:
                val, e_cat = get_yrs_tag_value(row['full_tag'])
                add_to_entities_dict(entities_dict, cat, {'entity': entity, cat.lower(): val, 'cat': e_cat})
            except ValueError:
                print(
                        f'Tag is neither year nor age: {row["full_tag"]}\nCheck: {file_path}\n'
                        f'{reconstructed_miu_text_with_tags}'
                )
                yml_handler.set_error_while_collecting_annotated_entities(row["full_tag"])
                return
        elif cat == 'TOPONYM':
            # Identify toponym
            place, uris_tag, list_of_uris, list_of_provinces = Toponyms().look_up_entity(entity)
            if notna(sub_cat):
                add_to_entities_dict(entities_dict, cat, {'entity': place, 'URI': uris_tag, 'cat': sub_cat})
            else:
                add_to_entities_dict(entities_dict, cat, {'entity': place, 'URI': uris_tag})
            if len(list_of_uris) == 0:
                path, uri = split(file_path)
                uri, ext = splitext(uri)
                LOGGER_TOPONYMS_UNKNOWN.info(f'{uri},{entity}')
            else:
                if len(list_of_provinces) > 0:
                    settlements_set.update(list_of_uris)
                    provinces_set.update(list_of_provinces)
                else:
                    # if there is no province URI that means the toponym is a province and should therefore be added
                    # to provinces and not settlements
                    provinces_set.update(list_of_uris)
                if len(list_of_uris) > 1:
                    # The toponym is ambiguous and matched multiple entries in our gazetteer
                    ambiguous_toponyms = True
        elif cat == 'PERSON' or cat == 'BOOK':
            if notna(sub_cat):
                add_to_entities_dict(entities_dict, cat, {'entity': entity, 'cat': sub_cat}, tag)
            else:
                print(
                    f'Tag is missing the sub-classification: {row["full_tag"]}\nCheck: {file_path}\n'
                    f'{reconstructed_miu_text_with_tags}'
                    )
                yml_handler.set_error_while_collecting_annotated_entities(row["full_tag"])
                add_to_entities_dict(entities_dict, cat, {'entity': entity, 'cat': 'X'}, tag)
        elif cat == 'ONOMASTIC':
            if tag.startswith('SHR') and entity.startswith('ب'):
                entity = entity[1:]
                add_to_entities_dict(entities_dict, cat, entity, tag)
            elif tag.startswith('NAS'):
                nas_dict['nas_' + str(nas_counter)] = entity
                nas_counter += 1
            else:
                add_to_entities_dict(entities_dict, cat, entity, tag)
            LOGGER_NASAB_KNOWN.info(f'{tag},{entity}')
        else:
            add_to_entities_dict(entities_dict, cat, entity, tag)

    if nas_dict != {}:
        if 'onomastics' in entities_dict.keys():
            entities_dict['onomastics']['nas'] = nas_dict
        else:
            entities_dict['onomastics'] = {'nas': nas_dict}

    if 'onomastics' in entities_dict.keys():
        # Sort dict by keys
        entities_dict['onomastics'] = dict(
                sorted(
                        [(k, list(v)) if isinstance(v, set) else (k, v) for k, v in
                         entities_dict.get('onomastics').items()],
                        key=itemgetter(0)
                )
        )

    # Generate edges
    if settlements_set:
        entities_dict['settlements'] = list(settlements_set)
        entities_dict['edges_settlements'] = [[a, b] for a, b in combinations(settlements_set, 2) if a != b]
    if provinces_set:
        entities_dict['provinces'] = list(provinces_set)
        entities_dict['edges_provinces'] = [[a, b] for a, b in combinations(provinces_set, 2) if a != b]

    # Extrapolate active period for the biographee
    # TODO dates_headings are not analysed yet
    if entities_dict.get('dates'):
        dates = entities_dict.get('dates')
        birth_date = [d.get('date') for d in dates if d.get('cat') == 'B']
        death_date = [d.get('date') for d in dates if d.get('cat') == 'D']
        alternative_dates = [d.get('date') for d in dates]
        age_at_death = None
        if entities_dict.get('ages') and [a.get('age') for a in entities_dict.get('ages') if a.get('cat') == 'D']:
            age_at_death = max([a.get('age') for a in entities_dict.get('ages') if a.get('cat') == 'D'])

        if death_date:
            entities_dict['max_date'] = max(death_date)
        elif birth_date:
            if age_at_death:
                entities_dict['max_date'] = min(birth_date) + age_at_death
            else:
                # TODO better approximation for active period
                entities_dict['max_date'] = min(birth_date) + 70
        elif alternative_dates:
            entities_dict['max_date'] = max(alternative_dates)

        if birth_date:
            entities_dict['min_date'] = min(birth_date)
        elif entities_dict.get('max_date'):
            if type(age_at_death) == 'int':
                entities_dict.get('max_date') - age_at_death
            else:
                # TODO better approximation for active period
                entities_dict['min_date'] = entities_dict.get('max_date') - 70

    # Entities, which are not listed in 'YAMLHandler.__attr_from_annotation' are ignored
    yml_handler.add_tagged_entities(entities_dict)
    if ambiguous_toponyms:
        yml_handler.set_ambiguous_toponyms()


def add_statistics_to_yml(
        df: DataFrame,
        yml_handler: YAMLHandler,
) -> None:
    tokens_without_punctuation_df = df['TOKENS'].loc[df['TOKENS'].notna() & ~df['TOKENS'].isin(UNICODE_PUNCT_CHARSET)]
    yml_handler.add_number_of_tokens(tokens_without_punctuation_df.count())  # Count() counts all non-na values



from typing import Optional

from os.path import split, splitext

from eis1600.markdown.markdown_patterns import HEADER_END_PATTERN, NORMALIZE_BIO_CHR_MD_PATTERN, \
    PAGE_TAG_SPLITTING_PARAGRAPH_PATTERN, SPACES_CROWD_PATTERN, NEWLINES_CROWD_PATTERN, POETRY_PATTERN, \
    SPACES_AFTER_NEWLINES_PATTERN, PAGE_TAG_ON_NEWLINE_MARKDOWN_PATTERN, BIO_CHR_TO_NEWLINE_PATTERN


def normalize_bio_chr_md(paragraph: str) -> str:
    md = NORMALIZE_BIO_CHR_MD_PATTERN.match(paragraph).group(0)
    if md == '$BIO_MAN$':
        return NORMALIZE_BIO_CHR_MD_PATTERN.sub('# $', paragraph)
    elif md == '$BIO_WOM$':
        return NORMALIZE_BIO_CHR_MD_PATTERN.sub('# $$', paragraph)
    elif md == '$BIO_REF$':
        return NORMALIZE_BIO_CHR_MD_PATTERN.sub('# $$$', paragraph)
    elif md == '$CHR_EVE$':
        return NORMALIZE_BIO_CHR_MD_PATTERN.sub('# @', paragraph)
    elif md == '$CHR_RAW$':
        return NORMALIZE_BIO_CHR_MD_PATTERN.sub('# @@@', paragraph)
    elif md == '@ RAW':
        return NORMALIZE_BIO_CHR_MD_PATTERN.sub('# @@@', paragraph)
    else:
        return paragraph


def convert_to_EIS1600TMP(infile: str, output_dir: Optional[str] = None, verbose: bool = False) -> None:
    """Coverts a file to EIS1600TMP for review process.

    Converts mARkdown, inProgress, completed file to light EIS1600TMP for the review process. Creates the file with the
    '.EIS1600TMP' extension.

    :param str infile: Path of the file to convert.
    :param str or None output_dir: Directory to write new file to (discontinued), optional.
    :param bool verbose: If True outputs a notification of the file which is currently processed, defaults to False.
    :return None:
    """
    if output_dir:
        path, uri = split(infile)
        uri, ext = splitext(uri)
        outfile = output_dir + '/' + uri + '.EIS1600TMP'
    else:
        path, ext = splitext(infile)
        outfile = path + '.EIS1600TMP'
        path, uri = split(infile)

    if verbose:
        print(f'Convert {uri} from mARkdown to EIS1600 file')

    with open(infile, 'r', encoding='utf8') as infile_h:
        text = infile_h.read()

    header_and_text = HEADER_END_PATTERN.split(text)
    header = header_and_text[0] + header_and_text[1]
    text = header_and_text[2][1:]  # Ignore second new line after #META#Header#End#

    if text[0:2] == '#\n':
        # Some texts start with a plain #, remove these
        text = text[2:]

    # fix
    text = text.replace('~\n', '\n')
    text = text.replace('\n~~', ' ')
    text = text.replace(' \n', '\n')

    # spaces
    text = SPACES_AFTER_NEWLINES_PATTERN.sub('\n', text)
    text = SPACES_CROWD_PATTERN.sub(' ', text)

    # fix poetry
    text = POETRY_PATTERN.sub(r'\1', text)

    # fix page tag on newlines
    text = PAGE_TAG_SPLITTING_PARAGRAPH_PATTERN.sub(r'\1 \2 \3', text)
    text = PAGE_TAG_ON_NEWLINE_MARKDOWN_PATTERN.sub(r' \1\n', text)
    text = SPACES_CROWD_PATTERN.sub(' ', text)

    # fix new lines
    text = text.replace('\n###', '\n\n###')
    text = text.replace('\n# ', '\n\n')
    text = NEWLINES_CROWD_PATTERN.sub('\n\n', text)

    text = text.split('\n\n')

    text_updated = []

    for paragraph in text:
        if paragraph.startswith('### '):
            paragraph = paragraph.replace('###', '#')
            if NORMALIZE_BIO_CHR_MD_PATTERN.match(paragraph):
                paragraph = normalize_bio_chr_md(paragraph)
            paragraph = BIO_CHR_TO_NEWLINE_PATTERN.sub(r'\1\n\2', paragraph)
        text_updated.append(paragraph)

    # reassemble text
    text = '\n\n'.join(text_updated)
    final = header + '\n\n' + text
    if final[-1] != '\n':
        final += '\n'

    with open(outfile, 'w', encoding='utf8') as outfile_h:
        outfile_h.write(final)



from sys import argv
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from glob import iglob

from tqdm.dask import TqdmCallback
from dask.dataframe import from_pandas
from numpy import nan
from pandas import concat, DataFrame, notna, read_csv

from eis1600.processing.preprocessing import get_yml_and_miu_df, get_tokens_and_tags
from eis1600.processing.postprocessing import merge_tagslists, reconstruct_miu_text_with_tags
from eis1600.repositories.repo import MIU_REPO, TOPO_REPO
from eis1600.toponym_descriptions.topod_extract_incomplete import PLACES_REGEX, TT_REGEX


def annotated_to_miu(row):
    infile = row['infile']
    outfile = TOPO_REPO + 'MIUs/' + row['MIU'] + '.EIS1600'
    with open(infile, 'r', encoding='utf-8') as miu_file_object:
        yml_handler, df = get_yml_and_miu_df(miu_file_object)

    text = ' '.join(df['TOKENS'].loc[df['TOKENS'].notna()].to_list())
    text_updated = text
    if all(notna(row['STATUS'])):
        for phrase in row['MODIFIABLE']:
            p = phrase.strip()
            tag = 'Q' + str(len(p.split())) + ' '
            text_updated = text_updated.replace(p, tag + p)

        if not 'NISBA' in row['STATUS']:
            yml_handler.set_reviewed('HRH')

    else:
        if PLACES_REGEX.search(text_updated):
            m = PLACES_REGEX.search(text_updated)
            while m:
                end = m.end()
                if len(TT_REGEX.findall(m.group(0))) >= 3:
                    p = m.group(0).strip()
                    tag = 'Q' + str(len(p.split())) + ' '
                    text_updated = text_updated.replace(p, tag + p)
                    m = PLACES_REGEX.search(text_updated, end + len(tag))
                else:
                    m = PLACES_REGEX.search(text_updated, end)

    tokens, tags = get_tokens_and_tags(text_updated)

    df.loc[df['TOKENS'].notna(), 'TPD_TAGS'] = tags
    df['TAGS_LISTS'] = df.apply(merge_tagslists, key='TPD_TAGS', axis=1)

    updated_text = reconstruct_miu_text_with_tags(df[['SECTIONS', 'TOKENS', 'TAGS_LISTS']])

    with open(outfile, 'w', encoding='utf-8') as outfile_h:
        outfile_h.write(str(yml_handler) + updated_text)


def get_infile_path(miu: str) -> str:
    author, text, version, uid = miu.split('.')
    return MIU_REPO + 'data/' + '/'.join([author, '.'.join([author, text])]) + '/MIUs/' + miu + '.EIS1600'


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to annotate onomastic information in gold-standard MIUs.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')

    args = arg_parser.parse_args()
    debug = args.debug

    sheets = iglob(TOPO_REPO + 'sheet_*.csv')
    sheets_df = DataFrame(None, columns=['MIU', 'MODIFIABLE', 'STATUS'])

    for sheet in sheets:
        tmp = read_csv(sheet, usecols=['MIU', 'MODIFIABLE', 'STATUS'])
        sheets_df = concat([sheets_df, tmp])

    sheets_df.dropna(how='all', inplace=True)
    sheets_df.drop_duplicates(inplace=True)
    sheets_df.dropna(inplace=True)

    sheets_df['STATUS'].loc[sheets_df['STATUS'].notna() & sheets_df['STATUS'].str.fullmatch('TOPOPONYM|CORRECT')] = 'TOPONYM'
    sheets_df['STATUS'].loc[sheets_df['STATUS'].notna() & sheets_df['STATUS'].str.fullmatch('ODD|INCOMPLETE')] = nan
    df_by_files = sheets_df.groupby('MIU').agg(list).reset_index()
    infiles = [get_infile_path(miu) for miu in df_by_files['MIU'].to_list()]

    df_by_files['infile'] = infiles

    ddf = from_pandas(df_by_files, npartitions=16)
    with TqdmCallback(desc="Status Toponym"):
        ddf.apply(annotated_to_miu, axis=1, meta=('apply', object)).compute()

    print('Done')



from threading import Lock
from typing import List, Optional

from camel_tools.disambig.mle import MLEDisambiguator
from camel_tools.tagger.default import DefaultTagger
from eis1600.helper.Singleton import singleton


@singleton
class Lemmatizer:
    def __init__(self) -> None:
        self.model = DefaultTagger(MLEDisambiguator.pretrained(), 'lex')
        self.lock = Lock()

    def get_lemmas(self, tokens: List[str], debug: Optional[bool] = False) -> List[str]:
        with self.lock:
            if debug:
                print('lemmatize')
            return self.model.tag(tokens)



class Month:
    """

    :param int month: month.
    :param int length: number of tokens which make up the date expression.
    """
    def __init__(self, month: int, length: int):
        self.month = month
        self.length = length

    def __eq__(self, other):
        return self.month == other.month and self.length == other.length

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return str(self.__dict__)

    def get_tag(self):
        # Use L because M tags are already used for misc
        tag = 'ÜL'
        tag += str(self.length)
        tag += 'X'

        month = str(self.month)
        while len(month) < 2:
            month = '0' + month

        tag += month

        return tag + 'L '



from pathlib import Path
from sys import argv
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from typing import Dict, List, Tuple

from p_tqdm import p_uimap

from json import dump
from datetime import datetime
from numpy import nan
from pandas import concat, DataFrame, notna, Series
import evaluate
from tensorflow import constant
from tensorflow.python.keras.metrics import MeanAbsoluteError, MeanAbsolutePercentageError

from eis1600.bio.md_to_bio import md_to_bio
from eis1600.dates.date_patterns import DATE_CATEGORIES_REPLACEMENTS, DATE_LABEL_DICT
from eis1600.dates.date_methods import date_annotate_miu_text
from eis1600.markdown.markdown_patterns import YEAR_PATTERN
from eis1600.model_evaluations.EvalResultsEncoder import EvalResultsEncoder
from eis1600.depricated.disassemble_reassemble_methods import get_yml_and_miu_df
from eis1600.repositories.repo import TRAINING_DATA_REPO, TRAINING_RESULTS_REPO


def reconstruct_automated_tag(row) -> str:
    return 'ÜY' + row['num_tokens'] + row['cat'] + row['written'] + 'Y'


def get_year_true_pred(row: Series) -> Series:
    """Get numerical values for ground-truth and prediction.

    :param Series row: row has values for 'written_true' and 'written_pred'.
    :return Series: Series of 'true' and 'pred'.
    """
    if notna(row['written_true']):
        v_true = int(row['written_true'])
    else:
        v_true = nan

    if notna(row['written_pred']) and row['written_pred'] != 'None':
        v_pred = int(row['written_pred'])
    else:
        v_pred = nan

    return Series([v_true, v_pred], index=['true', 'pred'])


def get_dates_true_and_pred(file: str) -> Tuple[DataFrame, Dict, Dict]:
    """Get ground-truth and prediction on labels and numerical values for MIU.

    :param str file: file path for MIU file.
    :param Dict bio_dict: BIO labels dictionary.
    :return Tuple[DataFrame, Dict, Dict]: DataFrame year contains two columns ('true' and 'pred'), the other two are
    dictionaries, one with the BIO labels derived from the ground-truth and the other with the BIO labels based on
    the predictions.
    """
    with open(file, 'r', encoding='utf-8') as miu_file_object:
        yml_handler, df = get_yml_and_miu_df(miu_file_object)

    # Extract the ground-truth annotated with EIS1600 tags
    s_notna = df['TAGS_LISTS'].loc[df['TAGS_LISTS'].notna()].apply(lambda tag_list: ','.join(tag_list))
    df_true = s_notna.str.extract(YEAR_PATTERN).dropna(how='all')
    dates = df_true.apply(reconstruct_automated_tag, axis=1)
    dates.name = 'DATES_TRUE'

    if not dates.empty:
        df = df.join(dates)
    else:
        df['DATES_TRUE'] = nan

    # Predict dates based on the text as EIS1600 tags
    df['DATES_PRED'] = date_annotate_miu_text(df[['TOKENS']], file)

    if any(df['DATES_PRED'].notna()):
        df_pred = df['DATES_PRED'].str.extract(YEAR_PATTERN).dropna(how='all')
    else:
        df_pred = DataFrame(columns=['written'])

    year = DataFrame(columns=['true', 'pred'])

    # Get information from the EIS1600 tags for predictions and ground-truth
    if not dates.empty:
        df_year = df_true[['written']].join(df_pred['written'], lsuffix='_true', rsuffix='_pred')
        year[['true', 'pred']] = df_year.apply(get_year_true_pred, axis=1)
    elif not df_pred.empty:
        df_year = df_pred[['written']].join(df_true['written'], rsuffix='_true', lsuffix='_pred')
        year[['true', 'pred']] = df_year.apply(get_year_true_pred, axis=1)

    year.dropna(inplace=True)

    # Parse EIS1600 tags to BIO tags for predictions and ground-truth
    bio_true = md_to_bio(
            df[['TOKENS', 'DATES_TRUE']],
            'DATES_TRUE',
            YEAR_PATTERN,
            'YY',
            DATE_LABEL_DICT,
            True,
            DATE_CATEGORIES_REPLACEMENTS
    )
    bio_pred = md_to_bio(
            df[['TOKENS', 'DATES_PRED']],
            'DATES_PRED',
            YEAR_PATTERN,
            'YY',
            DATE_LABEL_DICT,
            True,
            DATE_CATEGORIES_REPLACEMENTS
    )

    return year, bio_true, bio_pred


def eval_dates_entity_recognition_and_classification(truth: List[List[str]], predictions: List[List[str]]) -> Dict:
    """Evaluates predicted BIO-labels based on their ground-truth.

    Evaluates predicted BIO-labels based on their ground-truth, giving each of the following metrics for every class:
    precision, recall, F1-score, accuracy.

    :param List[List[str]] truth: A list of records, where each record is a list of the true BIO-tags for that text.
    :param List[List[str]] predictions: A list of records, where each record is a list of the predicted BIO-tags for
    that text.
    :return Dict: Dict containing the evaluation results: precision, recall, F1 for each class.
    """
    all_metrics = evaluate.load("seqeval").compute(predictions=predictions, references=truth)

    return all_metrics


def eval_year(year: DataFrame) -> Dict:
    """Evaluates the predictions of the numerical value of all recognized date with MAPE and MAE.

    :param DataFrame year: DataFrame with two columns, column 'true' with the ground-truth values and column 'pred'
    with the predicted values.
    :return Dict: Dict containing the evaluation results: Mean Average Percentage Error and Mean Average Error.
    """
    all_metrics = {}
    truth = constant(year['true'].astype('float32'))
    predictions = constant(year['pred'].astype('float32'))

    mape = MeanAbsolutePercentageError()
    mape.update_state(y_true=truth, y_pred=predictions)
    mae_no_nan = MeanAbsoluteError()
    mae_no_nan.update_state(y_true=truth, y_pred=predictions)

    all_metrics['mape'] = mape.result().numpy()
    all_metrics['mae'] = mae_no_nan.result().numpy()

    return all_metrics


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description='''Script to annotate onomastic information in gold-standard MIUs.'''
    )
    arg_parser.add_argument('-D', '--debug', action='store_true')

    args = arg_parser.parse_args()
    debug = args.debug

    with open(TRAINING_DATA_REPO + 'gold_standard.txt', 'r', encoding='utf-8') as fh:
        files_txt = fh.read().splitlines()

    infiles = [TRAINING_DATA_REPO + 'gold_standard/' + file for file in files_txt if Path(
            TRAINING_DATA_REPO + 'gold_standard/' + file).exists()]

    res = []
    if debug:
        for i, file in enumerate(infiles):
            print(i, file)
            res.append(get_dates_true_and_pred(file))
    else:
        res += p_uimap(get_dates_true_and_pred, infiles)

        years, truth, predictions = zip(*res)

        with open(TRAINING_DATA_REPO + 'dates_truth.json', 'w', encoding='utf-8') as fh:
            dump(truth, fh, indent=4, ensure_ascii=False)
        with open(TRAINING_DATA_REPO + 'dates_predictions.json', 'w', encoding='utf-8') as fh:
            dump(predictions, fh, indent=4, ensure_ascii=False)

    truth = []
    predictions = []
    year_df = DataFrame(columns=['true', 'pred'])

    for year, bio_true, bio_pred in res:
        truth.append(bio_true['ner_classes'])
        predictions.append(bio_pred['ner_classes'])
        if not year.empty:
            year_df = concat([year_df, year])

    all_metrics = {
            'model': 'dates-rule-based'
    }
    all_metrics.update(eval_dates_entity_recognition_and_classification(truth, predictions))
    all_metrics.update(eval_year(year_df))

    with open(TRAINING_RESULTS_REPO + 'dates/result-' + datetime.now().strftime('%Y%m%d-%H%M%S') + '.json', 'w',
              encoding='utf-8') as fh:
        dump(all_metrics, fh, cls=EvalResultsEncoder)

    print('Done')



from typing import List

from eis1600.helper.ar_normalization import denormalize_list
from importlib_resources import files
from pandas import read_csv

from eis1600.helper.Singleton import singleton

path = files('eis1600.gazetteers.data').joinpath('spelling_gazetteer.csv')


@singleton
class Spellings:
    __tot = None
    __denormalized = None
    __regex = ''

    def __init__(self) -> None:
        try:
            df = read_csv(path)
        except Exception as e:
            print(f'Check if the {self.__class__.__name__} gazetteer is correctly formatted with commas and not tabs.')
        df['NGRAM'] = df['NGRAM'].astype('uint8')
        df['CATEGORY'] = df['CATEGORY'].astype('category')

        sorted_df = df.sort_values(by=['NGRAM'], ascending=False)
        Spellings.__tot = sorted_df['VALUE'].to_list()
        denormalized = []
        [denormalized.extend(denormalize_list(t)) for t in Spellings.__tot]
        Spellings.__denormalized = denormalized
        Spellings.__regex = r'(?P<spelling> [وب]?(?:ال)?(?:' + '|'.join(denormalized) + ')(?:ها)?)'

    @staticmethod
    def get_denormalized_list() -> List[str]:
        return Spellings.__denormalized

    @staticmethod
    def get_regex() -> str:
        return Spellings.__regex



from pathlib import Path
from sys import argv
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from pandas import DataFrame

from eis1600.corpus_analysis.text_methods import get_text_as_list_of_mius
from eis1600.helper.CheckFileEndingActions import CheckFileEndingEIS1600OrEIS1600TMPAction
from eis1600.paragraphs.paragraph_methods import redefine_paragraphs
from eis1600.repositories.repo import NEW_PARAGRAPHS_REPO, POETRY_TEST_RES_REPO, TEXT_REPO


def main():
    arg_parser = ArgumentParser(
            prog=argv[0], formatter_class=RawDescriptionHelpFormatter,
            description=''
    )
    arg_parser.add_argument(
            'input', type=str, nargs='1',
            help='EIS1600 or EIS1600TMP file to process',
            action=CheckFileEndingEIS1600OrEIS1600TMPAction
    )

    args = arg_parser.parse_args()

    # TODO ATM it only works with a single file so you can test how the output will look like. When this routine is
    #  ready for use, it should be part of the incorporate new files or ids_insert_or_update routine
    infile = args.input

    meta_data_header, mius_list = get_text_as_list_of_mius(infile)
    data = []
    mius = []
    x = 0
    for i, tup in enumerate(mius_list[x:]):
        uid, miu_as_text, analyse_flag = tup
        print(i + x, uid)
        poetry_tests_list, miu = redefine_paragraphs(uid, miu_as_text)
        mius.append(miu)
        data += [(uid, dict_per_paragraph['text'], dict_per_paragraph['label'], dict_per_paragraph['score']) for
               dict_per_paragraph in poetry_tests_list]

    outfile = infile.replace(TEXT_REPO, NEW_PARAGRAPHS_REPO)
    path_parts = outfile.split('/')
    Path('/'.join(path_parts[:-1])).mkdir(exist_ok=True, parents=True)
    with open(outfile, 'w', encoding='utf-8') as fh:
        fh.write(meta_data_header + ''.join(mius))
    poetry_test_res_path = infile.replace(TEXT_REPO, POETRY_TEST_RES_REPO).replace('EIS1600', 'csv')
    path_parts = poetry_test_res_path.split('/')
    Path('/'.join(path_parts[:-1])).mkdir(exist_ok=True, parents=True)
    DataFrame(data, columns=['uid', 'text', 'meter', 'score']).to_csv(poetry_test_res_path)



from typing import Tuple, Union
from argparse import ArgumentTypeError


def parse_range(arg: str) -> Tuple[int, Union[int, None]]:
    try:
        i, j = arg.split(",")
        i = int(i) - 1 if i else 0
        j = int(j) if j else None
        return i, j
    except ValueError:
        raise ArgumentTypeError("range must be i,j with both i and j being integers")




from numpy import nan

from pandas import DataFrame, Series
from typing import Match, Optional

from openiti.helper.ara import normalize_ara_heavy

from eis1600.dates.Date import Date
from eis1600.dates.date_patterns import DATE_CATEGORIES_NOR, DATE_CATEGORY_PATTERN, DATE_PATTERN, MONTHS_NOR, \
    ONES_NOR, TEN_NOR, HUNDRED_NOR, THOUSAND_NOR, ONES_HINDI, SANA_PATTERN
from eis1600.processing.preprocessing import get_tokens_and_tags
from eis1600.yml.YAMLHandler import YAMLHandler


def parse_year(m: Match[str]) -> (int, int):
    """

    :param m:
    :return:
    :raise ValueError:
    """
    year = 0
    year_digits = 0
    length = len(m.group('sana').split())  # (?P<sana>سنة|عام|في حدود)
    if m.group('ones'):
        year += ONES_NOR.get(normalize_ara_heavy(m.group('ones')))
        length += 1
    if m.group('ten'):
        year += TEN_NOR.get(normalize_ara_heavy(m.group('ten')))
        length += 1
    if m.group('hundred'):
        year += HUNDRED_NOR.get(normalize_ara_heavy(m.group('hundred')))
        length += len(m.group('hundred').split())
    if m.group('thousand'):
        year += THOUSAND_NOR.get(normalize_ara_heavy(m.group('thousand')))
        length += 1
    if m.group('digits'):
        digits = m.group('digits')
        if digits.isdigit():
            year_digits = int(digits)
        else:
            multiplier = [1, 10, 100, 1000]
            c = len(digits)
            while c:
                year_digits += ONES_HINDI.get(digits[c - 1]) * multiplier[c - 1]

        length += len(m.group('digits_str').strip().split())

        if year == 0:
            year = year_digits
        elif year != year_digits:
            raise ValueError(
                    f"Date recognition: parsed value and given value are at odds. Check {m.group(0)}\n"
                    f"given: {year_digits}\n"
                    f"parsed: {year}"
            )

    return year, length


def get_dates_headings(yml_handler: YAMLHandler) -> None:
    """Checks the headings for date statements and if a such a statement is found, it is converted into a tag and
    added to the yml header.

    :param YAMLHandler yml_handler: arabic text.
    """
    headings = yml_handler.headings
    for key, val in headings:
        if DATE_PATTERN.search(val):
            m = DATE_PATTERN.search(val)
            year, length = parse_year(m)
            yml_handler.add_date_headings(year)


def tag_dates_fulltext(text: str) -> str:
    """Inserts EIS1600 date tags in the arabic text and returns the text with the tags.

    This is a rule-based model to recognize, classify and parse date phrases in Arabic texts.
    The text is returned with all recognized dates annotated with EIS1600 tags for dates.
    :param str text: arabic text.
    :return str: arabic text with EIS1600 tags for the recognized and classified dates.
    """
    text_updated = text
    # Regex that recognizes date phrases (and unfortunately age phrases as well)
    m = DATE_PATTERN.search(text_updated)
    while m:
        # While date phrases are recognized in the text
        if m.group('year') and not SANA_PATTERN.fullmatch(m.group('year')):
            # Check if phrase is an actual date. Date phrases give the years after the word sana. In
            # contrast, age phrases give the years before the word sana.
            year = 0
            length = 1
            month = None
            day = 0
            weekday = None
            # Length is one because sana is definitely recognized
            try:
                year, length = parse_year(m)
            except ValueError as e:
                raise e
            else:
                # Date classification
                if DATE_CATEGORY_PATTERN.search(m.group('context')):
                    last = DATE_CATEGORY_PATTERN.findall(m.group('context'))[-1]
                    date_category = DATE_CATEGORIES_NOR.get(normalize_ara_heavy(last))
                else:
                    date_category = 'X'

                # Parsing of other information from the date phrase as day of the week, day of the month, month
                # Currently not of interest
                # if m.group('weekday'):
                #     weekday = WEEKDAYS_NOR.get(normalize_ara_heavy(m.group('weekday')))
                # if m.group('day_ones'):
                #     day += DAY_ONES_NOR.get(normalize_ara_heavy(m.group('day_ones')))
                # if m.group('day_ten'):
                #     day += DAY_TEN_NOR.get(normalize_ara_heavy(m.group('day_ten')))
                # if m.group('month'):
                #     month_str = normalize_ara_heavy(m.group('month'))
                #     month = MONTHS_NOR.get(month_str)
                # else:
                #     mm = MONTH_PATTERN.search(m[0])
                #     if mm:
                #         month_str = mm[0]
                #         month = MONTHS.get(month_str)

                # if day == 0:
                #     day = None
                if year == 0:
                    year = None

                # Get extracted information bundled as date object and insert EIS600 tag for date into the text
                date = Date(year, length, date_category)
                pos = m.start('sana')
                text_updated = text_updated[:pos] + date.get_tag() + text_updated[pos:]

                # Recognize next date phrase
                m = DATE_PATTERN.search(text_updated, m.end('sana') + len(date.get_tag()))
        else:
            # Recognize next date phrase
            m = DATE_PATTERN.search(text_updated, m.end('sana'))

    return text_updated


def date_annotate_miu_text(ner_df: DataFrame, file: str, yml: Optional[YAMLHandler] = None) -> Series:
    """Annotate dates in the headings and in the MIU text, returns a list of tag per token.

    :param DataFrame ner_df: df containing the 'TOKENS' column.
    :param str file: file name.
    :param YAMLHandler yml: yml_header to collect date tags in.
    :return Series: List of date tags per token, which can be added as additional column to the df.
    """
    if yml:
        get_dates_headings(yml)

    df = ner_df.mask(ner_df == '', None)
    tokens = df['TOKENS'].dropna()
    ar_text = ' '.join(tokens)

    try:
        tagged_text = tag_dates_fulltext(ar_text)
    except ValueError as e:
        print(e)
        print(f'Check {file}')

        return Series([nan] * len(df))
    else:
        ar_tokens, tags = get_tokens_and_tags(tagged_text)
        df.loc[df['TOKENS'].notna(), 'DATE_TAGS'] = tags

        return df['DATE_TAGS']



from logging import FileHandler, Formatter, getLogger, INFO, Logger

FORMATTER = Formatter('%(message)s')


def setup_logger(name: str, log_file: str, level=INFO, formatter=FORMATTER) -> Logger:

    handler = FileHandler(log_file, 'w', encoding='utf-8')
    handler.setFormatter(formatter)

    logger = getLogger(name)
    logger.setLevel(level)
    logger.addHandler(handler)

    return logger



from sys import exit
from os.path import isfile
from urllib import request
from urllib.error import URLError


def download_text_selection(text_repo: str) -> str:
    """Download the Google Sheet which keeps tracks of prepared texts.

    :param text_repo: TEXT_REPO cannot be import due to circular import, hence as param.
    :return str: return path to the downloaded CSV file.
    """
    csv_path = text_repo + '_EIS1600 - Text Selection - Serial Source Test - EIS1600_AutomaticSelectionForReview.csv'
    print(
            'Download latest version of "_EIS1600 - Text Selection - Serial Source Test - '
            'EIS1600_AutomaticSelectionForReview" from Google Spreadsheets'
    )
    latest_csv = "https://docs.google.com/spreadsheets/d/e/2PACX-1vR60MrlXJjtrd3bid1CR3xK5Pv" \
                 "-aUz1qWEfHfowU1DPsh6RZBvbtW2mA-83drzboIS1fxZdsDO-ny0r/pub?gid=2075964223&single=true&output=csv"
    request.urlcleanup()
    try:
        response = request.urlopen(latest_csv)
    except URLError:
        print('Could not download latest version, fall back to locally saved version')
        if isfile(csv_path):
            return csv_path
        else:
            print(f'There is no local version of that file, please download it manually and save it here:\n{csv_path}')
            exit()
    else:
        lines = [line.decode('utf-8') for line in response.readlines()]
        with open(csv_path, 'w', encoding='utf-8') as csv_fh:
            csv_fh.writelines(lines)

        print('Saved as csv in ' + text_repo + '\n')

        return csv_path



from typing import Dict


def dict_to_yaml(d: Dict, level: int) -> str:
    intend = '    ' * level
    yaml = ''
    for key2, val2 in d.items():
        if isinstance(val2, dict):
            val_str = dict_to_yaml(val2, level + 1)
            yaml += intend + f'- {key2} :\n{val_str}\n'
        else:
            yaml += intend + f'- {key2} : {str(val2)}\n'

    return yaml[:-1]



from eis1600.markdown.markdown_patterns import MIU_UID_TAG_AND_TEXT_SAME_LINE_PATTERN, \
    NEW_LINE_BUT_NO_EMPTY_LINE_PATTERN, \
    NEW_LINE_INSIDE_PARAGRAPH_NOT_POETRY_PATTERN, PARAGRAPH_TAG_MISSING, EMPTY_PARAGRAPH_CHECK_PATTERN, \
    SIMPLE_MARKDOWN, MISSING_DIRECTIONALITY_TAG_PATTERN, SPAN_ELEMENTS, TEXT_START_PATTERN, TILDA_HICKUPS_PATTERN
from eis1600.markdown.markdown_patterns import SIMPLE_MARKDOWN_TEXT_START_PATTERN


def check_file_for_mal_formatting(infile: str, content: str):
    if not TEXT_START_PATTERN.match(content) \
            or PARAGRAPH_TAG_MISSING.search(content) \
            or SIMPLE_MARKDOWN.search(content) \
            or NEW_LINE_BUT_NO_EMPTY_LINE_PATTERN.search(content) \
            or MIU_UID_TAG_AND_TEXT_SAME_LINE_PATTERN.search(content) \
            or TILDA_HICKUPS_PATTERN.search(content) \
            or NEW_LINE_INSIDE_PARAGRAPH_NOT_POETRY_PATTERN.search(content) \
            or EMPTY_PARAGRAPH_CHECK_PATTERN.search(content) \
            or SPAN_ELEMENTS.search(content) \
            or MISSING_DIRECTIONALITY_TAG_PATTERN.search(content):
        # Poetry is still to messed up, do not bother with it for now
        # or POETRY_ATTACHED_AFTER_PAGE_TAG.search(content):
        error = ''
        if not TEXT_START_PATTERN.match(content):
            error += '\n * Text does not start with a MIU tag, check if the preface is tagged as PARATEXT.'
        if PARAGRAPH_TAG_MISSING.search(content):
            error += '\n * There are missing paragraph tags.'
        if SIMPLE_MARKDOWN.search(content):
            error += '\n * There is simple mARkdown left.'
        if NEW_LINE_BUT_NO_EMPTY_LINE_PATTERN.search(content):
            error += '\n * There are elements missing the double newline (somewhere the emtpy line is missing).'
        if MIU_UID_TAG_AND_TEXT_SAME_LINE_PATTERN.search(content):
            error += '\n * There is text on the same line as the start of a biography, fix it by running ' \
                     f'`ids_insert_or_update` on {infile}'
        if TILDA_HICKUPS_PATTERN.search(content):
            error += '\n * There is this pattern with tildes: `~\\n~`.'
        if NEW_LINE_INSIDE_PARAGRAPH_NOT_POETRY_PATTERN.search(content):
            error += '\n * There is a single newline inside a paragraph (somewhere the emtpy line is missing).'
        if EMPTY_PARAGRAPH_CHECK_PATTERN.search(content):
            error += '\n * There are empty paragraphs in the text.'
        if SPAN_ELEMENTS.search(content):
            error += '\n * There are span elements in the text.'
        if MISSING_DIRECTIONALITY_TAG_PATTERN.search(content):
            error += '\n * There are missing direction tags at the beginning of paragraphs, fix it by running ' \
                     f'`ids_insert_or_update` on {infile}'
        # if POETRY_ATTACHED_AFTER_PAGE_TAG.search(content):
        #     error += '\n * There is poetry attached to a PageTag (there should be a linebreak instead).'

        raise ValueError(
                f'Correct the following errors\n'
                f'open -a kate {infile}\n'
                f'kate {infile}\n'
                f'{error}\n\n'
                f'And now run\n'
                f'ids_insert_or_update {infile}\n'
                f'Check if everything is working with\n'
                f'check_formatting {infile}\n'
        )


def check_formatting(infile: str):
    with open(infile, 'r', encoding='utf8') as text:
        header_text = text.read().split('#META#Header#End#\n\n')

        try:
            check_file_for_mal_formatting(infile, header_text[1])
        except ValueError:
            raise


def check_formatting_before_update_ids(infile: str, content: str):
    if not TEXT_START_PATTERN.match(content) and not SIMPLE_MARKDOWN_TEXT_START_PATTERN.match(content) \
            or NEW_LINE_BUT_NO_EMPTY_LINE_PATTERN.search(content) \
            or TILDA_HICKUPS_PATTERN.search(content) \
            or NEW_LINE_INSIDE_PARAGRAPH_NOT_POETRY_PATTERN.search(content) \
            or EMPTY_PARAGRAPH_CHECK_PATTERN.search(content) \
            or SPAN_ELEMENTS.search(content):

        error = ''
        if not TEXT_START_PATTERN.match(content) and not SIMPLE_MARKDOWN_TEXT_START_PATTERN.match(content):
            error += '\n * Text does not start with Header or PARATEXT, check if the preface is tagged.'
        if NEW_LINE_BUT_NO_EMPTY_LINE_PATTERN.search(content):
            error += '\n * There are elements missing the double newline (somewhere the emtpy line is missing).'
        if TILDA_HICKUPS_PATTERN.search(content):
            error += '\n * There is this pattern with tildes: `~\\n~`.'
        if NEW_LINE_INSIDE_PARAGRAPH_NOT_POETRY_PATTERN.search(content):
            error += '\n * There is a single newline inside a paragraph (somewhere the emtpy line is missing).'
        if EMPTY_PARAGRAPH_CHECK_PATTERN.search(content):
            error += '\n * There are empty paragraphs in the text.'
        if SPAN_ELEMENTS.search(content):
            error += '\n * There are span elements in the text.'

        raise ValueError(
                f'Correct the following errors\n'
                f'open -a kate {infile}\n'
                f'kate {infile}\n'
                f'{error}\n\n'
                f'And now run\n'
                f'ids_insert_or_update {infile}\n'
                f'Check if everything is working with\n'
                f'check_formatting {infile}\n'
        )



from argparse import Action
from os.path import isdir, isfile, splitext


class CheckFileEndingMARKdownAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and isfile(input_arg):
            filepath, fileext = splitext(input_arg)
            if fileext not in ['.mARkdown', '.inProgess', '.completed']:
                parser.error('You need to input a mARkdown file')
            else:
                setattr(namespace, self.dest, input_arg)
        else:
            setattr(namespace, self.dest, None)


class CheckFileEndingEIS1600MIUAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and isfile(input_arg):
            filepath, fileext = splitext(input_arg)
            if fileext != '.EIS1600' and filepath[-1:-12].isdigit():
                parser.error('Input must be a single MIU file')
            else:
                setattr(namespace, self.dest, input_arg)
        else:
            setattr(namespace, self.dest, None)


class CheckFileEndingEIS1600TextAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and isfile(input_arg):
            filepath, fileext = splitext(input_arg)
            if fileext != '.EIS1600' and not filepath[-1:-12].isdigit():
                parser.error('Input must be a single MIU file')
            else:
                setattr(namespace, self.dest, input_arg)
        else:
            setattr(namespace, self.dest, None)


class CheckFileEndingEIS1600OrEIS1600TMPAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and isfile(input_arg):
            filepath, fileext = splitext(input_arg)
            if not fileext.startswith('.EIS1600'):
                parser.error('You need to input an EIS1600 or EIS1600TMP file')
            else:
                setattr(namespace, self.dest, input_arg)
        else:
            setattr(namespace, self.dest, None)


class CheckFileEndingEIS1600OrIDsAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        if input_arg and isfile(input_arg):
            filepath, fileext = splitext(input_arg)
            if fileext != '.IDs' and fileext != '.EIS1600':
                parser.error('You need to input an IDs file or a single MIU file')
            else:
                setattr(namespace, self.dest, input_arg)
        else:
            setattr(namespace, self.dest, None)


class CheckIsDirAction(Action):
    def __call__(self, parser, namespace, input_arg, option_string=None):
        input_arg = input_arg[0]
        if input_arg and isdir(input_arg):
            setattr(namespace, self.dest, input_arg)
        else:
            print('You need to specify a valid path to the directory holding the files which have been annotated')
            raise IOError
